# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for optimizing your media based on media mix models."""
import functools
from typing import Optional, Tuple, Union
from absl import logging
import jax
import jax.numpy as jnp
from scipy import optimize

from lightweight_mmm import lightweight_mmm
from lightweight_mmm import preprocessing


@functools.partial(
    jax.jit,
    static_argnames=("media_mix_model", "media_input_shape", "target_scaler",
                     "media_scaler"))
def _objective_function(extra_features: jnp.ndarray,
                        media_mix_model: lightweight_mmm.LightweightMMM,
                        media_input_shape: Tuple[int,
                                                 int], media_gap: Optional[int],
                        target_scaler: Optional[preprocessing.CustomScaler],
                        media_scaler: preprocessing.CustomScaler,
                        geo_ratio: jnp.array,
                        seed: Optional[int],
                        media_values: jnp.ndarray) -> jnp.float64:
  """Objective function to calculate the sum of all predictions of the model.

  Args:
    extra_features: Extra features the model requires for prediction.
    media_mix_model: Media mix model to use. Must have a predict method to be
      used.
    media_input_shape: Input shape of the data required by the model to get
      predictions. This is needed since optimization might flatten some arrays
      and they need to be reshaped before running new predictions.
    media_gap: Media data gap between the end of training data and the start of
      the out of sample media given. Eg. if 100 weeks of data were used for
      training and prediction starts 2 months after training data finished we
      need to provide the 8 weeks missing between the training data and the
      prediction data so data transformations (adstock, carryover, ...) can take
      place correctly.
    target_scaler: Scaler that was used to scale the target before training.
    media_scaler: Scaler that was used to scale the media data before training.
    geo_ratio: The ratio to split channel media across geo. Should sum up to 1
      for each channel and should have shape (c, g).
    seed: Seed to use for PRNGKey during sampling. For replicability run
      this function and any other function that gets predictions with the same
      seed.
    media_values: Media values required by the model to run predictions.

  Returns:
    The negative value of the sum of all predictions.
  """
  if hasattr(media_mix_model, "n_geos") and media_mix_model.n_geos > 1:
    media_values = geo_ratio * jnp.expand_dims(media_values, axis=-1)
  media_values = jnp.tile(
      media_values / media_input_shape[0], reps=media_input_shape[0])
  # Distribute budget of each channels across time.
  media_values = jnp.reshape(a=media_values, newshape=media_input_shape)
  media_values = media_scaler.transform(media_values)
  return -jnp.sum(
      media_mix_model.predict(
          media=media_values.reshape(media_input_shape),
          extra_features=extra_features,
          media_gap=media_gap,
          target_scaler=target_scaler,
          seed=seed).mean(axis=0))


@jax.jit
def _budget_constraint(media: jnp.ndarray,
                       prices: jnp.ndarray,
                       budget: jnp.ndarray) -> jnp.float64:
  """Calculates optimization constraint to keep spend equal to the budget.

  Args:
    media: Array with the values of the media for this iteration.
    prices: Prices of each media channel at any given time.
    budget: Total budget of the optimization.

  Returns:
    The result from substracting the total spending and the budget.
  """
  media = media.reshape((-1, len(prices)))
  return jnp.sum(media * prices) - budget


def _get_lower_and_upper_bounds(
    media: jnp.ndarray,
    n_time_periods: int,
    lower_pct: jnp.ndarray,
    upper_pct: jnp.ndarray,
    media_scaler: Optional[preprocessing.CustomScaler] = None
) -> optimize.Bounds:
  """Gets the lower and upper bounds for optimisation based on historic data.

  It creates an upper bound based on a percentage above the mean value on
  each channel and a lower bound based on a relative decrease of the mean
  value.

  Args:
    media: Media data to get historic mean.
    n_time_periods: Number of time periods to optimize for. If model is built on
      weekly data, this would be the number of weeks ahead to optimize.
    lower_pct: Relative percentage decrease from the mean value to consider as
      new lower bound.
    upper_pct: Relative percentage increase from the mean value to consider as
      new upper bound.
    media_scaler: Scaler that was used to scale the media data before training.

  Returns:
    A list of tuples with the lower and upper bound for each media channel.
  """
  if media.ndim == 3:
    lower_pct = jnp.expand_dims(lower_pct, axis=-1)
    upper_pct = jnp.expand_dims(upper_pct, axis=-1)

  mean_data = media.mean(axis=0)
  lower_bounds = jnp.maximum(mean_data * (1 - lower_pct), 0)
  upper_bounds = mean_data * (1 + upper_pct)

  if media_scaler:
    lower_bounds = media_scaler.inverse_transform(lower_bounds)
    upper_bounds = media_scaler.inverse_transform(upper_bounds)

  if media.ndim == 3:
    lower_bounds = lower_bounds.sum(axis=-1)
    upper_bounds = upper_bounds.sum(axis=-1)

  return optimize.Bounds(lb=lower_bounds * n_time_periods,
                         ub=upper_bounds * n_time_periods)


def _generate_starting_values(
    n_time_periods: int, media: jnp.ndarray,
    media_scaler: preprocessing.CustomScaler,
    budget: Union[float, int],
    prices: jnp.ndarray,
) -> jnp.ndarray:
  """Generates starting values based on historic allocation and budget.

  In order to make a comparison we can take the allocation of the last
  `n_time_periods` and scale it based on the given budget. Given this, one can
  compare how this initial values (based on average historic allocation) compare
  to the output of the optimisation in terms of sales/KPI.

  Args:
    n_time_periods: Number of time periods the optimization will be done with.
    media: Historic media data the model was trained with.
    media_scaler: Scaler that was used to scale the media data before training.
    budget: Total budget to allocate during the optimization time.
    prices: An array with shape (n_media_channels,) for the cost of each media
      channel unit.

  Returns:
    An array with the starting value for each media channel for the
      optimization.
  """
  previous_allocation = media.mean(axis=0) * n_time_periods
  if media_scaler:  # Scale before sum as geo scaler has shape (c, g).
    previous_allocation = media_scaler.inverse_transform(previous_allocation)

  if media.ndim == 3:
    previous_allocation = previous_allocation.sum(axis=-1)

  avg_spend_per_channel = previous_allocation * prices
  pct_spend_per_channel = avg_spend_per_channel / avg_spend_per_channel.sum()
  budget_per_channel = budget * pct_spend_per_channel
  media_unit_per_channel = budget_per_channel / prices
  return media_unit_per_channel


def find_optimal_budgets(
    n_time_periods: int,
    media_mix_model: lightweight_mmm.LightweightMMM,
    budget: Union[float, int],
    prices: jnp.ndarray,
    extra_features: Optional[jnp.ndarray] = None,
    media_gap: Optional[jnp.ndarray] = None,
    target_scaler: Optional[preprocessing.CustomScaler] = None,
    media_scaler: Optional[preprocessing.CustomScaler] = None,
    bounds_lower_pct: Union[float, jnp.ndarray] = .2,
    bounds_upper_pct: Union[float, jnp.ndarray] = .2,
    max_iterations: int = 200,
    solver_func_tolerance: float = 1e-06,
    solver_step_size: float = 1.4901161193847656e-08,
    seed: Optional[int] = None) -> optimize.OptimizeResult:
  """Finds the best media allocation based on MMM model, prices and a budget.

  Args:
    n_time_periods: Number of time periods to optimize for. If model is built on
      weekly data, this would be the number of weeks ahead to optimize.
    media_mix_model: Media mix model to use for the optimization.
    budget: Total budget to allocate during the optimization time.
    prices: An array with shape (n_media_channels,) for the cost of each media
      channel unit.
    extra_features: Extra features needed for the model to predict.
    media_gap: Media data gap between the end of training data and the start of
      the out of sample media given. Eg. if 100 weeks of data were used for
      training and prediction starts 8 weeks after training data finished we
      need to provide the 8 weeks missing between the training data and the
      prediction data so data transformations (adstock, carryover, ...) can take
      place correctly.
    target_scaler: Scaler that was used to scale the target before training.
    media_scaler: Scaler that was used to scale the media data before training.
    bounds_lower_pct: Relative percentage decrease from the mean value to
      consider as new lower bound.
    bounds_upper_pct: Relative percentage increase from the mean value to
      consider as new upper bound.
    max_iterations: Number of max iterations to use for the SLSQP scipy
      optimizer. Default is 200.
    solver_func_tolerance: Precision goal for the value of the prediction in
      the stopping criterion. Maps directly to scipy's `ftol`. Intended only
      for advanced users. For more details see:
      https://docs.scipy.org/doc/scipy/reference/optimize.minimize-slsqp.html#optimize-minimize-slsqp.
    solver_step_size: Step size used for numerical approximation of the
      Jacobian. Maps directly to scipy's `eps`. Intended only for advanced
      users. For more details see:
      https://docs.scipy.org/doc/scipy/reference/optimize.minimize-slsqp.html#optimize-minimize-slsqp.
    seed: Seed to use for PRNGKey during sampling. For replicability run
      this function and any other function that gets predictions with the same
      seed.

  Returns:
    solution: OptimizeResult object containing the results of the optimization.
    kpi_without_optim: Predicted target based on original allocation proportion
    among channels from the historical data.
    starting_values: Budget Allocation based on original allocation proportion
    and the given total budget.
  """
  if not hasattr(media_mix_model, "media"):
    raise ValueError(
        "The passed model has not been trained. Please fit the model before "
        "running optimization.")
  jax.config.update("jax_enable_x64", True)

  if isinstance(bounds_lower_pct, float):
    bounds_lower_pct = jnp.repeat(a=bounds_lower_pct, repeats=len(prices))
  if isinstance(bounds_upper_pct, float):
    bounds_upper_pct = jnp.repeat(a=bounds_upper_pct, repeats=len(prices))

  bounds = _get_lower_and_upper_bounds(
      media=media_mix_model.media,
      n_time_periods=n_time_periods,
      lower_pct=bounds_lower_pct,
      upper_pct=bounds_upper_pct,
      media_scaler=media_scaler)
  if jnp.sum(bounds.lb * prices) > budget:
    logging.warning(
        "Budget given is smaller than the lower bounds of the constraints for "
        "optimization. This will lead to faulty optimization. Please either "
        "increase the budget or change the lower bound by increasing the "
        "percentage decrease with the `bounds_lower_pct` parameter.")
  if jnp.sum(bounds.ub * prices) < budget:
    logging.warning(
        "Budget given is larger than the upper bounds of the constraints for "
        "optimization. This will lead to faulty optimization. Please either "
        "reduce the budget or change the upper bound by increasing the "
        "percentage increase with the `bounds_upper_pct` parameter.")

  starting_values = _generate_starting_values(
      n_time_periods=n_time_periods,
      media=media_mix_model.media,
      media_scaler=media_scaler,
      budget=budget,
      prices=prices,
  )
  if not media_scaler:
    media_scaler = preprocessing.CustomScaler(multiply_by=1, divide_by=1)
  if media_mix_model.n_geos == 1:
    geo_ratio = 1.0
  else:
    average_per_time = media_mix_model.media.mean(axis=0)
    geo_ratio = average_per_time / jnp.expand_dims(
        average_per_time.sum(axis=-1), axis=-1)
  media_input_shape = (n_time_periods, *media_mix_model.media.shape[1:])
  partial_objective_function = functools.partial(
      _objective_function, extra_features, media_mix_model,
      media_input_shape, media_gap,
      target_scaler, media_scaler, geo_ratio, seed)
  solution = optimize.minimize(
      fun=partial_objective_function,
      x0=starting_values,
      bounds=bounds,
      method="SLSQP",
      jac="3-point",
      options={
          "maxiter": max_iterations,
          "disp": True,
          "ftol": solver_func_tolerance,
          "eps": solver_step_size,
      },
      constraints={
          "type": "eq",
          "fun": _budget_constraint,
          "args": (prices, budget)
      })

  kpi_without_optim = _objective_function(extra_features=extra_features,
                                          media_mix_model=media_mix_model,
                                          media_input_shape=media_input_shape,
                                          media_gap=media_gap,
                                          target_scaler=target_scaler,
                                          media_scaler=media_scaler,
                                          seed=seed,
                                          geo_ratio=geo_ratio,
                                          media_values=starting_values)
  logging.info("KPI without optimization: %r", -1 * kpi_without_optim.item())
  logging.info("KPI with optimization: %r", -1 * solution.fun)

  jax.config.update("jax_enable_x64", False)
  # TODO(yukaabe): Create an object to contain the results of this function.
  return solution, kpi_without_optim, starting_values



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for optimize_media."""
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
import jax
import jax.numpy as jnp
import numpy as np

from lightweight_mmm import lightweight_mmm
from lightweight_mmm import optimize_media
from lightweight_mmm import preprocessing


class OptimizeMediaTest(parameterized.TestCase):

  @classmethod
  def setUpClass(cls):
    super(OptimizeMediaTest, cls).setUpClass()
    cls.national_mmm = lightweight_mmm.LightweightMMM()
    cls.national_mmm.fit(
        media=jnp.ones((50, 5)),
        target=jnp.ones(50),
        media_prior=jnp.ones(5) * 50,
        number_warmup=2,
        number_samples=2,
        number_chains=1)
    cls.geo_mmm = lightweight_mmm.LightweightMMM()
    cls.geo_mmm.fit(
        media=jnp.ones((50, 5, 3)),
        target=jnp.ones((50, 3)),
        media_prior=jnp.ones(5) * 50,
        number_warmup=2,
        number_samples=2,
        number_chains=1)

  def setUp(self):
    super().setUp()
    self.mock_minimize = self.enter_context(
        mock.patch.object(optimize_media.optimize, "minimize", autospec=True))

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          model_name="national_mmm",
          geo_ratio=1),
      dict(
          testcase_name="geo",
          model_name="geo_mmm",
          geo_ratio=np.tile(0.33, reps=(5, 3)))
  ])
  def test_objective_function_generates_correct_value_type_and_sign(
      self, model_name, geo_ratio):

    mmm = getattr(self, model_name)
    extra_features = mmm._extra_features
    time_periods = 10

    kpi_predicted = optimize_media._objective_function(
        extra_features=extra_features,
        media_mix_model=mmm,
        media_input_shape=(time_periods, *mmm.media.shape[1:]),
        media_gap=None,
        target_scaler=None,
        media_scaler=preprocessing.CustomScaler(),
        media_values=jnp.ones(mmm.n_media_channels) * time_periods,
        geo_ratio=geo_ratio,
        seed=10)

    self.assertIsInstance(kpi_predicted, jax.Array)
    self.assertLessEqual(kpi_predicted, 0)
    self.assertEqual(kpi_predicted.shape, ())

  @parameterized.named_parameters([
      dict(
          testcase_name="zero_output",
          media=np.ones(9),
          prices=np.array([1, 2, 3]),
          budget=18,
          expected_value=0),
      dict(
          testcase_name="negative_output",
          media=np.ones(9),
          prices=np.array([1, 2, 3]),
          budget=20,
          expected_value=-2),
      dict(
          testcase_name="positive_output",
          media=np.ones(9),
          prices=np.array([1, 2, 3]),
          budget=16,
          expected_value=2),
      dict(
          testcase_name="bigger_array",
          media=np.ones(18),
          prices=np.array([2, 2, 2]),
          budget=36,
          expected_value=0),
  ])
  def test_budget_constraint(self, media, prices, budget, expected_value):
    generated_value = optimize_media._budget_constraint(
        media=media, prices=prices, budget=budget)

    self.assertEqual(generated_value, expected_value)

  @parameterized.named_parameters([
      dict(
          testcase_name="national_media_scaler",
          model_name="national_mmm"),
      dict(
          testcase_name="geo_media_scaler",
          model_name="geo_mmm")
  ])
  def test_find_optimal_budgets_with_scaler_optimize_called_with_right_params(
      self, model_name):

    mmm = getattr(self, model_name)
    media_scaler = preprocessing.CustomScaler(divide_operation=jnp.mean)
    media_scaler.fit(2 * jnp.ones((10, *mmm.media.shape[1:])))
    optimize_media.find_optimal_budgets(
        n_time_periods=15,
        media_mix_model=mmm,
        budget=30,
        prices=jnp.ones(mmm.n_media_channels),
        target_scaler=None,
        media_scaler=media_scaler)

    _, call_kwargs = self.mock_minimize.call_args_list[0]
    # 15 weeks at 1.2 gives us 12. and 18. bounds times 2 (scaler) 24. and 36.
    np.testing.assert_array_almost_equal(call_kwargs["bounds"].lb,
                                         np.repeat(24., repeats=5) * mmm.n_geos,
                                         decimal=3)
    np.testing.assert_array_almost_equal(call_kwargs["bounds"].ub,
                                         np.repeat(36., repeats=5) * mmm.n_geos,
                                         decimal=3)
    # We only added scaler with divide operation so we only expectec x2 in
    # the divide_by parameter.
    np.testing.assert_array_almost_equal(call_kwargs["fun"].args[5].divide_by,
                                         2 * jnp.ones(mmm.media.shape[1:]),
                                         decimal=3)
    np.testing.assert_array_almost_equal(call_kwargs["fun"].args[5].multiply_by,
                                         jnp.ones(mmm.media.shape[1:]),
                                         decimal=3)

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          model_name="national_mmm"),
      dict(
          testcase_name="geo",
          model_name="geo_mmm")
  ])
  def test_find_optimal_budgets_without_scaler_optimize_called_with_right_params(
      self, model_name):

    mmm = getattr(self, model_name)
    optimize_media.find_optimal_budgets(
        n_time_periods=15,
        media_mix_model=mmm,
        budget=30,
        prices=jnp.ones(mmm.n_media_channels),
        target_scaler=None,
        media_scaler=None)

    _, call_kwargs = self.mock_minimize.call_args_list[0]
    # 15 weeks at 1.2 gives us 18. bounds
    np.testing.assert_array_almost_equal(
        call_kwargs["bounds"].lb,
        np.repeat(12., repeats=5) * mmm.n_geos,
        decimal=3)
    np.testing.assert_array_almost_equal(
        call_kwargs["bounds"].ub,
        np.repeat(18., repeats=5) * mmm.n_geos,
        decimal=3)

    np.testing.assert_array_almost_equal(
        call_kwargs["fun"].args[5].divide_by,
        jnp.ones(mmm.n_media_channels),
        decimal=3)
    np.testing.assert_array_almost_equal(
        call_kwargs["fun"].args[5].multiply_by,
        jnp.ones(mmm.n_media_channels),
        decimal=3)

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          model_name="national_mmm"),
      dict(
          testcase_name="geo",
          model_name="geo_mmm")
  ])
  def test_predict_called_with_right_args(self, model_name):
    mmm = getattr(self, model_name)

    optimize_media.find_optimal_budgets(
        n_time_periods=15,
        media_mix_model=mmm,
        budget=30,
        prices=jnp.ones(mmm.n_media_channels),
        target_scaler=None,
        media_scaler=None)

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          model_name="national_mmm"),
      dict(
          testcase_name="geo",
          model_name="geo_mmm")
  ])
  def test_budget_lower_than_constraints_warns_user(self, model_name):
    mmm = getattr(self, model_name)
    expected_warning = (
        "Budget given is smaller than the lower bounds of the constraints for "
        "optimization. This will lead to faulty optimization. Please either "
        "increase the budget or change the lower bound by increasing the "
        "percentage decrease with the `bounds_lower_pct` parameter.")

    with self.assertLogs(level="WARNING") as context_manager:
      optimize_media.find_optimal_budgets(
          n_time_periods=5,
          media_mix_model=mmm,
          budget=1,
          prices=jnp.ones(mmm.n_media_channels))
    self.assertEqual(f"WARNING:absl:{expected_warning}",
                     context_manager.output[0])

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          model_name="national_mmm"),
      dict(
          testcase_name="geo",
          model_name="geo_mmm")
  ])
  def test_budget_higher_than_constraints_warns_user(self, model_name):
    mmm = getattr(self, model_name)
    expected_warning = (
        "Budget given is larger than the upper bounds of the constraints for "
        "optimization. This will lead to faulty optimization. Please either "
        "reduce the budget or change the upper bound by increasing the "
        "percentage increase with the `bounds_upper_pct` parameter.")

    with self.assertLogs(level="WARNING") as context_manager:
      optimize_media.find_optimal_budgets(
          n_time_periods=5,
          media_mix_model=mmm,
          budget=2000,
          prices=jnp.ones(5))
    self.assertEqual(f"WARNING:absl:{expected_warning}",
                     context_manager.output[0])

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          model_name="national_mmm",
          expected_len=3),
      dict(
          testcase_name="geo",
          model_name="geo_mmm",
          expected_len=3)
  ])
  def test_find_optimal_budgets_has_right_output_length_datatype(
      self, model_name, expected_len):

    mmm = getattr(self, model_name)
    media_scaler = preprocessing.CustomScaler(divide_operation=jnp.mean)
    media_scaler.fit(2 * jnp.ones((10, *mmm.media.shape[1:])))
    results = optimize_media.find_optimal_budgets(
        n_time_periods=15,
        media_mix_model=mmm,
        budget=30,
        prices=jnp.ones(mmm.n_media_channels),
        target_scaler=None,
        media_scaler=media_scaler)
    self.assertLen(results, expected_len)
    self.assertIsInstance(results[1], jax.Array)
    self.assertIsInstance(results[2], jax.Array)

  @parameterized.named_parameters([
      dict(
          testcase_name="national_prices",
          model_name="national_mmm",
          prices=np.array([1., 0.8, 1.2, 1.5, 0.5]),
      ),
      dict(
          testcase_name="national_ones",
          model_name="national_mmm",
          prices=np.ones(5),
      ),
      dict(
          testcase_name="geo_prices",
          model_name="geo_mmm",
          prices=np.array([1., 0.8, 1.2, 1.5, 0.5]),
      ),
      dict(
          testcase_name="geo_ones",
          model_name="geo_mmm",
          prices=np.ones(5),
      ),
  ])
  def test_generate_starting_values_calculates_correct_values(
      self, model_name, prices):
    mmm = getattr(self, model_name)
    n_time_periods = 10
    budget = mmm.n_media_channels * n_time_periods
    starting_values = optimize_media._generate_starting_values(
        n_time_periods=10,
        media_scaler=None,
        media=mmm.media,
        budget=budget,
        prices=prices,
    )

    # Given that data is all ones, starting values will be equal to prices.
    np.testing.assert_array_almost_equal(
        starting_values, jnp.repeat(n_time_periods, repeats=5))

if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A simple and lightweight library for Media Mix Modelling.

Simple usage of this class goes as following:

```
mmm = lightweight_mmm.LightweightMMM()
mmm.fit(media=media_data,
        extra_features=extra_features,
        media_prior=costs,
        target=target,
        number_samples=1000,
        number_chains=2)

# For obtaining media contribution percentage and ROI
predictions, media_contribution_hat_pct, roi_hat = mmm.get_posterior_metrics()

# For running predictions on unseen data
mmm.predict(media=media_data_test, extra_features=extra_features_test)
```
"""

import collections
import dataclasses
import functools
import itertools
import logging
import numbers
from typing import Any, Callable, Dict, Mapping, MutableMapping, Optional, Sequence, Tuple, Union

from absl import logging
import immutabledict
import jax
import jax.numpy as jnp
import numpy as np
import numpyro
from numpyro import distributions as dist
from numpyro import infer

from lightweight_mmm import models
from lightweight_mmm import preprocessing
from lightweight_mmm import utils

Prior = Union[
    dist.Distribution,
    Dict[str, float],
    Sequence[float],
    float
]

_NAMES_TO_MODEL_TRANSFORMS = immutabledict.immutabledict({
    "hill_adstock": models.transform_hill_adstock,
    "adstock": models.transform_adstock,
    "carryover": models.transform_carryover
})
_MODEL_FUNCTION = models.media_mix_model


def _compare_equality_for_lmmm(item_1: Any, item_2: Any) -> bool:
  """Compares two items for equality.

  Helper function for the __eq__ method of LightweightmMM. First checks if items
  are strings or lists of strings (it's okay if empty lists compare True), then
  uses jnp.array_equal if the items are jax.numpy.DeviceArray or other related
  sequences, and uses items' __eq__ otherwise.

  Note: this implementation does not cover every possible data structure, but
  it does cover all the data structures seen in attributes used by
  LightweightMMM. Sometimes the DeviceArray is hidden in the value of a
  MutableMapping, hence the recursion.

  Args:
    item_1: First item to be compared.
    item_2: Second item to be compared.

  Returns:
    Boolean for whether item_1 equals item_2.
  """

  # This is pretty strict but LMMM classes don't need to compare equal unless
  # they are exact copies.
  if type(item_1) != type(item_2):
    is_equal = False
  elif isinstance(item_1, str):
    is_equal = item_1 == item_2
  elif isinstance(item_1, (jax.Array, np.ndarray, Sequence)):
    if all(isinstance(x, str) for x in item_1) and all(
        isinstance(x, str) for x in item_2):
      is_equal = item_1 == item_2
    else:
      is_equal = jnp.array_equal(item_1, item_2, equal_nan=True)
  elif isinstance(item_1, MutableMapping):
    is_equal = all([
        _compare_equality_for_lmmm(item_1[x], item_2[x])
        for x in item_1.keys() | item_2.keys()
    ])
  else:
    is_equal = item_1 == item_2

  return is_equal


class NotFittedModelError(Exception):
  pass


@dataclasses.dataclass(unsafe_hash=True, eq=False)
class LightweightMMM:
  """Lightweight Media Mix Modelling wrapper for bayesian models.

  The currently available models are the following:
   - hill_adstock
   - adstock
   - carryover

  It also offers the necessary utilities for calculating media contribution and
  media ROI based on models' results.

  Attributes:
    trace: Sampling trace of the bayesian model once fitted.
    n_media_channels: Number of media channels the model was trained with.
    n_geos: Number of geos for geo models or 1 for national models.
    model_name: Name of the model.
    media: The media data the model is trained on. Usefull for a variety of
      insights post model fitting.
    media_names: Names of the media channels passed at fitting time.
    custom_priors: The set of custom priors the model was trained with. An empty
      dictionary if none were passed.
  """
  model_name: str = "hill_adstock"
  n_media_channels: int = dataclasses.field(init=False, repr=False)
  n_geos: int = dataclasses.field(init=False, repr=False)
  media: jnp.DeviceArray = dataclasses.field(
      init=False, repr=False, hash=False, compare=True)
  media_names: Sequence[str] = dataclasses.field(
      init=False, repr=False, hash=False, compare=True)
  trace: Dict[str, jnp.DeviceArray] = dataclasses.field(
      init=False, repr=False, hash=False, compare=False)
  custom_priors: MutableMapping[str, Prior] = dataclasses.field(
      init=False, repr=False, hash=False, compare=True)
  _degrees_seasonality: int = dataclasses.field(init=False, repr=False)
  _weekday_seasonality: bool = dataclasses.field(init=False, repr=False)
  _media_prior: jnp.DeviceArray = dataclasses.field(
      init=False, repr=False, hash=False, compare=True)
  _extra_features: jnp.DeviceArray = dataclasses.field(
      init=False, repr=False, hash=False, compare=True)
  _target: jnp.DeviceArray = dataclasses.field(
      init=False, repr=False, hash=False, compare=True)
  _train_media_size: int = dataclasses.field(
      init=False, repr=False, hash=True, compare=False)
  _mcmc: numpyro.infer.MCMC = dataclasses.field(
      init=False, repr=False, hash=False, compare=False)

  def __post_init__(self):
    if self.model_name not in _NAMES_TO_MODEL_TRANSFORMS:
      raise ValueError("Model name passed not valid. Please use any of the"
                       "following: 'hill_adstock', 'adstock', 'carryover'.")
    self._model_function = _MODEL_FUNCTION
    self._model_transform_function = _NAMES_TO_MODEL_TRANSFORMS[self.model_name]
    self._prior_names = models.MODEL_PRIORS_NAMES.union(
        models.TRANSFORM_PRIORS_NAMES[self.model_name])

  def __eq__(self, other: Any) -> bool:
    """Equality method for LightweightMMMM.

    We need a special method here to handle a couple of issues. First, some of
    the attributes for LightweightMMM are arrays, which contain multiple values
    and cannot be evaluated with the default __eq__ method. Second, some
    attributes are initially undefined and only get values after fitting a
    model. The latter is dealt with within this function, and the former within
    the helper function _compare_equality_for_lmmm().

    Args:
      other: Dataclass to compare against.

    Returns:
      Boolean for whether self == other; NotImplemented if other is not a
      LightweightMMM.
    """
    if not isinstance(other, LightweightMMM):
      return NotImplemented

    def _create_list_of_attributes_to_compare(
        mmm_instance: Any) -> Sequence[str]:
      all_attributes_that_can_be_compared = sorted(
          [x.name for x in dataclasses.fields(mmm_instance) if x.compare])
      attributes_which_have_been_instantiated = [
          x for x in all_attributes_that_can_be_compared
          if hasattr(mmm_instance, x)
      ]
      return attributes_which_have_been_instantiated

    self_attributes = _create_list_of_attributes_to_compare(self)
    other_attributes = _create_list_of_attributes_to_compare(other)

    return all(
        _compare_equality_for_lmmm(getattr(self, a1), getattr(other, a2))
        for a1, a2 in itertools.zip_longest(self_attributes, other_attributes))

  def _preprocess_custom_priors(
      self,
      custom_priors: Dict[str, Prior]) -> MutableMapping[str, Prior]:
    """Preprocesses the user input custom priors to Numpyro distributions.

    If numpyro distributions are given they remains untouched, however if any
    other option is passed, it is passed to the default distribution to alter
    its constructor values.

    Args:
      custom_priors: Mapping of the name of the prior to its custom value.

    Returns:
      A mapping of names to numpyro distributions based on user input and
        default values.
    """
    default_priors = {
        **models._get_default_priors(),
        **models._get_transform_default_priors()[self.model_name]
    }
    # Checking that the key is contained in custom_priors has already been done
    # at this point in the fit function.
    for prior_name in custom_priors:
      if isinstance(custom_priors[prior_name], numbers.Number):
        custom_priors[prior_name] = default_priors[prior_name].__class__(
            custom_priors[prior_name])
      elif (isinstance(custom_priors[prior_name], collections.abc.Sequence) and
            not isinstance(custom_priors[prior_name], str)):
        custom_priors[prior_name] = default_priors[prior_name].__class__(
            *custom_priors[prior_name])
      elif isinstance(custom_priors[prior_name], dict):
        custom_priors[prior_name] = default_priors[prior_name].__class__(
            **custom_priors[prior_name])
      elif not isinstance(custom_priors[prior_name], dist.Distribution):
        raise ValueError(
            "Priors given must be a Numpyro distribution or one of the "
            "following to fit in the constructor of our default Numpyro "
            "distribution. It could be given as args or kwargs as long as it "
            "is the correct format for such object. Please refer to our "
            "documentation on custom priors to know more.")
    return custom_priors

  def fit(
      self,
      media: jnp.ndarray,
      media_prior: jnp.ndarray,
      target: jnp.ndarray,
      extra_features: Optional[jnp.ndarray] = None,
      degrees_seasonality: int = 2,
      seasonality_frequency: int = 52,
      weekday_seasonality: bool = False,
      media_names: Optional[Sequence[str]] = None,
      number_warmup: int = 1000,
      number_samples: int = 1000,
      number_chains: int = 2,
      target_accept_prob: float = .85,
      init_strategy: Callable[[Mapping[Any, Any], Any],
                              jnp.ndarray] = numpyro.infer.init_to_median,
      custom_priors: Optional[Dict[str, Prior]] = None,
      seed: Optional[int] = None) -> None:
    """Fits MMM given the media data, extra features, costs and sales/KPI.

    For detailed information on the selected model please refer to its
    respective function in the models.py file.

    Args:
      media: Media input data. Media data must have either 2 dims for national
        model or 3 for geo models.
      media_prior: Costs of each media channel. The number of cost values must
        be equal to the number of media channels.
      target: Target KPI to use, like for example sales.
      extra_features: Other variables to add to the model.
      degrees_seasonality: Number of degrees to use for seasonality. Default is
        2.
      seasonality_frequency: Frequency of the time period used. Default is 52 as
        in 52 weeks per year.
      weekday_seasonality: In case of daily data, also estimate seven weekday
        parameters.
      media_names: Names of the media channels passed.
      number_warmup: Number of warm up samples. Default is 1000.
      number_samples: Number of samples during sampling. Default is 1000.
      number_chains: Number of chains to sample. Default is 2.
      target_accept_prob: Target acceptance probability for step size in the
        NUTS sampler. Default is .85.
      init_strategy: Initialization function for numpyro NUTS. The available
        options can be found in
        https://num.pyro.ai/en/stable/utilities.html#initialization-strategies.
        Default is numpyro.infer.init_to_median.
      custom_priors: The custom priors we want the model to take instead of the
        default ones. Refer to the full documentation on custom priors for
        details.
      seed: Seed to use for PRNGKey during training. For better replicability
        run all different trainings with the same seed.
    """
    if media.ndim not in (2, 3):
      raise ValueError(
          "Media data must have either 2 dims for national model or 3 for geo "
          "models.")
    if media.ndim == 3 and media_prior.ndim == 1:
      media_prior = jnp.expand_dims(media_prior, axis=-1)

    if media.shape[1] != len(media_prior):
      raise ValueError("The number of data channels provided must match the "
                       "number of cost values.")
    if media.min() < 0:
      raise ValueError("Media values must be greater or equal to zero.")

    if custom_priors:
      not_used_custom_priors = set(custom_priors.keys()).difference(
          self._prior_names)
      if not_used_custom_priors:
        raise ValueError(
            "The following passed custom priors dont have a match in the model."
            " Please double check the names have been written correctly: %s" %
            not_used_custom_priors)
      custom_priors = self._preprocess_custom_priors(
          custom_priors=custom_priors)
      geo_custom_priors = set(custom_priors.keys()).intersection(
          models.GEO_ONLY_PRIORS)
      if media.ndim == 2 and geo_custom_priors:
        raise ValueError(
            "The given data is for national models but custom_prior contains "
            "priors for the geo version of the model. Please either remove geo "
            "priors for national model or pass media data with geo dimension.")
    else:
      custom_priors = {}

    if weekday_seasonality and seasonality_frequency == 52:
      logging.warn("You have chosen daily seasonality and frequency 52 "
                   "(weekly), please check you made the right seasonality "
                   "choices.")

    if extra_features is not None:
      extra_features = jnp.array(extra_features)

    if seed is None:
      seed = utils.get_time_seed()

    train_media_size = media.shape[0]
    kernel = numpyro.infer.NUTS(
        model=self._model_function,
        target_accept_prob=target_accept_prob,
        init_strategy=init_strategy)

    mcmc = numpyro.infer.MCMC(
        sampler=kernel,
        num_warmup=number_warmup,
        num_samples=number_samples,
        num_chains=number_chains)
    mcmc.run(
        rng_key=jax.random.PRNGKey(seed),
        media_data=jnp.array(media),
        extra_features=extra_features,
        target_data=jnp.array(target),
        media_prior=jnp.array(media_prior),
        degrees_seasonality=degrees_seasonality,
        frequency=seasonality_frequency,
        transform_function=self._model_transform_function,
        weekday_seasonality=weekday_seasonality,
        custom_priors=custom_priors)

    self.custom_priors = custom_priors
    if media_names is not None:
      self.media_names = media_names
    else:
      self.media_names = [f"channel_{i}" for i in range(media.shape[1])]
    self.n_media_channels = media.shape[1]
    self.n_geos = media.shape[2] if media.ndim == 3 else 1
    self._media_prior = media_prior
    self.trace = mcmc.get_samples()
    self._number_warmup = number_warmup
    self._number_samples = number_samples
    self._number_chains = number_chains
    self._target = target
    self._train_media_size = train_media_size
    self._degrees_seasonality = degrees_seasonality
    self._seasonality_frequency = seasonality_frequency
    self._weekday_seasonality = weekday_seasonality
    self.media = media
    self._extra_features = extra_features
    self._mcmc = mcmc
    logging.info("Model has been fitted")

  def print_summary(self) -> None:
    """Calls print_summary function from numpyro to print parameters summary.
    """
    # TODO(): add name selection for print.
    self._mcmc.print_summary()

  @functools.partial(
      jax.jit,
      static_argnums=(0,),
      static_argnames=("degrees_seasonality", "weekday_seasonality",
                       "transform_function", "model"))
  def _predict(
      self,
      rng_key: jnp.ndarray,
      media_data: jnp.ndarray,
      extra_features: Optional[jnp.ndarray],
      media_prior: jnp.ndarray,
      degrees_seasonality: int, frequency: int,
      transform_function: Callable[[Any], jnp.ndarray],
      weekday_seasonality: bool,
      model: Callable[[Any], None],
      posterior_samples: Dict[str, jnp.ndarray],
      custom_priors: Dict[str, Prior]
      ) -> Dict[str, jnp.ndarray]:
    """Encapsulates the numpyro.infer.Predictive function for predict method.

    It serves as a helper jitted function for running predictions.

    Args:
      rng_key: A jax.random.PRNGKey.
      media_data: Media array for needed for the model to run predictions.
      extra_features: Extra features for needed for the model to run.
      media_prior: Cost prior used for training the model.
      degrees_seasonality: Number of degrees for the seasonality.
      frequency: Frequency of the seasonality.
      transform_function: Media transform function to use within the model.
      weekday_seasonality: Allow daily weekday estimation.
      model: Numpyro model to use for numpyro.infer.Predictive.
      posterior_samples: Mapping of the posterior samples.
      custom_priors: The custom priors we want the model to take instead of the
        default ones. Refer to the full documentation on custom priors for
        details.

    Returns:
      The predictions for the given data.
    """
    return infer.Predictive(
        model=model, posterior_samples=posterior_samples)(
            rng_key=rng_key,
            media_data=media_data,
            extra_features=extra_features,
            media_prior=media_prior,
            target_data=None,
            degrees_seasonality=degrees_seasonality,
            frequency=frequency,
            transform_function=transform_function,
            custom_priors=custom_priors,
            weekday_seasonality=weekday_seasonality)

  def predict(
      self,
      media: jnp.ndarray,
      extra_features: Optional[jnp.ndarray] = None,
      media_gap: Optional[jnp.ndarray] = None,
      target_scaler: Optional[preprocessing.CustomScaler] = None,
      seed: Optional[int] = None
  ) -> jnp.ndarray:
    """Runs the model to obtain predictions for the given input data.

    Predictions returned are distributions, if point estimates are desired one
    can calculate those based on the given distribution.

    Args:
      media: Media array for needed for the model to run predictions.
      extra_features: Extra features for needed for the model to run.
      media_gap: Media data gap between the end of training data and the start
        of the out of sample media given. Eg. if 100 weeks of data were used for
        training and prediction starts 2 months after training data finished we
        need to provide the 8 weeks missing between the training data and the
        prediction data so data transformations (adstock, carryover, ...) can
        take place correctly.
      target_scaler: Scaler that was used to scale the target before training.
      seed: Seed to use for PRNGKey during sampling. For replicability run
        this function and any other function that utilises predictions with the
        same seed.

    Returns:
      Predictions for the given media and extra features at a given date index.

    Raises:
      NotFittedModelError: When the model has not been fitted before running
        predict.
    """
    if not hasattr(self, "trace"):
      raise NotFittedModelError("Need to fit the model before running "
                                "predictions.")
    if media_gap is not None:
      if media.ndim != media_gap.ndim:
        raise ValueError("Original media data and media gap must have the same "
                         "number of dimensions.")
      if media.ndim > 1 and media.shape[1] != media_gap.shape[1]:
        raise ValueError("Media gap must have the same numer of media channels"
                         "as the original media data.")
      previous_media = jnp.concatenate(arrays=[self.media, media_gap], axis=0)
      if extra_features is not None:
        previous_extra_features = jnp.concatenate(
            arrays=[
                self._extra_features,
                jnp.zeros((media_gap.shape[0], *self._extra_features.shape[1:]))
            ],
            axis=0)
    else:
      previous_media = self.media
      previous_extra_features = self._extra_features

    full_media = jnp.concatenate(arrays=[previous_media, media], axis=0)
    if extra_features is not None:
      full_extra_features = jnp.concatenate(
          arrays=[previous_extra_features, extra_features], axis=0)
    else:
      full_extra_features = None
    if seed is None:
      seed = utils.get_time_seed()
    prediction = self._predict(
        rng_key=jax.random.PRNGKey(seed=seed),
        media_data=full_media,
        extra_features=full_extra_features,
        media_prior=jnp.array(self._media_prior),
        degrees_seasonality=self._degrees_seasonality,
        frequency=self._seasonality_frequency,
        weekday_seasonality=self._weekday_seasonality,
        transform_function=self._model_transform_function,
        model=self._model_function,
        custom_priors=self.custom_priors,
        posterior_samples=self.trace)["mu"][:, previous_media.shape[0]:]
    if target_scaler:
      prediction = target_scaler.inverse_transform(prediction)

    return prediction

  def reduce_trace(self, nsample: int = 100, seed: int = 0) -> None:
    """Reduces the samples in `trace` to speed up `predict` and optimize.

    Please note this step is not reversible. Only do this after you have
    investigated convergence of the model.

    Args:
      nsample: Target number of samples.
      seed: Random seed for down sampling.

    Raises:
      ValueError: if `nsample` is too big.
    """
    ntrace = len(self.trace["sigma"])
    if ntrace < nsample:
      raise ValueError("nsample is bigger than the actual posterior samples")
    key = jax.random.PRNGKey(seed)
    samples = jax.random.choice(key, ntrace, (nsample,), replace=False)
    for name in self.trace.keys():
      self.trace[name] = self.trace[name][samples]
    logging.info("Reduction is complete")

  def get_posterior_metrics(
      self,
      unscaled_costs: Optional[jnp.ndarray] = None,
      cost_scaler: Optional[preprocessing.CustomScaler] = None,
      target_scaler: Optional[preprocessing.CustomScaler] = None
  ) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """It estimates the media contribution percentage and ROI of each channel.

    If data was scaled prior to training then the target and costs scalers need
    to be passed to this function to correctly calculate media contribution
    percentage and ROI in the unscaled space.

    Args:
      unscaled_costs: Optionally you can pass new costs to get these set of
        metrics. If None, the costs used for training will be used for
        calculating ROI.
      cost_scaler: Scaler that was used to scale the cost data before training.
        It is ignored if 'unscaled_costs' is provided.
      target_scaler: Scaler that was used to scale the target before training.

    Returns:
      media_contribution_hat_pct: The average media contribution percentage for
      each channel.
      roi_hat: The return on investment of each channel calculated as its
      contribution divided by the cost.

    Raises:
      NotFittedModelError: When the this method is called without the model
        being trained previously.
    """
    if not hasattr(self, "trace"):
      raise NotFittedModelError(
          "LightweightMMM has not been fitted and cannot run estimations. "
          "Please first fit the model.")
    if unscaled_costs is None and not cost_scaler:
      logging.warning(
          "Unscaled cost data or cost scaler were not given and  "
          "therefore unscaling wont be applied to calculcate contribution"
          " and ROI. If data was not scaled prior to training "
          "please ignore this warning.")
    if not target_scaler:
      logging.warning("Target scaler was not given and unscaling of the target "
                      "will not occur. If your target was not scaled prior to "
                      "training you can ignore this warning.")
    if unscaled_costs is None:
      if cost_scaler:
        unscaled_costs = cost_scaler.inverse_transform(self._media_prior)
      else:
        unscaled_costs = self._media_prior

    if self.media.ndim == 3:
      # cost shape (channel, geo) -> add a new axis to (channel, geo, sample)
      unscaled_costs = unscaled_costs = unscaled_costs[:, :, jnp.newaxis]
      # reshape cost to (sample, channel, geo)
      unscaled_costs = jnp.einsum("cgs->scg", unscaled_costs)

    # get the scaled posterior prediction
    posterior_pred = self.trace["mu"]
    if target_scaler:
      unscaled_posterior_pred = target_scaler.inverse_transform(posterior_pred)
    else:
      unscaled_posterior_pred = posterior_pred

    if self.media.ndim == 2:
      # s for samples, t for time, c for media channels
      einsum_str = "stc, sc -> sc"
    elif self.media.ndim == 3:
      # s for samples, t for time, c for media channels, g for geo
      einsum_str = "stcg, scg -> scg"

    media_contribution = jnp.einsum(einsum_str, self.trace["media_transformed"],
                                    jnp.squeeze(self.trace["coef_media"]))

    # aggregate posterior_pred across time:
    sum_scaled_prediction = jnp.sum(posterior_pred, axis=1)
    # aggregate unscaled_posterior_pred across time:
    sum_unscaled_prediction = jnp.sum(unscaled_posterior_pred, axis=1)

    if self.media.ndim == 2:
      # add a new axis to represent channel:(sample,) -> (sample,channel)
      sum_scaled_prediction = sum_scaled_prediction[:, jnp.newaxis]
      sum_unscaled_prediction = sum_unscaled_prediction[:, jnp.newaxis]

    elif self.media.ndim == 3:
      # add a new axis to represent channel:(sample,geo) -> (sample,geo,channel)
      # note: the total prediction value stays the same for all channels
      sum_scaled_prediction = sum_scaled_prediction[:, jnp.newaxis, :]
      # add a new axis to represent channel:(sample,geo) -> (sample,geo,channel)
      # note: the total prediction value stays the same for all channels
      sum_unscaled_prediction = sum_unscaled_prediction[:, :, jnp.newaxis]
      # reshape the array (sample,geo,channel) -> (sample,channel,geo)
      sum_unscaled_prediction = jnp.einsum("sgc->scg", sum_unscaled_prediction)

    # media contribution pct = media contribution / prediction
    # for geo level model:
    # media_contribution shape (sample, channel, geo)
    # sum_scaled_prediction shape (sample, channel, geo)
    # -> media_contribution_hat shape (sample, channel, geo)
    media_contribution_hat = media_contribution / sum_scaled_prediction

    # media roi = unscaled prediction * media contribution pct / unscaled costs
    # for geo leve model:
    # sum_unscaled_prediction shape (sample, channel, geo)
    # media_contribution_hat shape (sample, channel, geo)
    # unscaled_costs shape (sample, channel, geo)
    # -> roi_hat shape (sample, channel, geo)
    roi_hat = sum_unscaled_prediction * media_contribution_hat / unscaled_costs

    return media_contribution_hat, roi_hat



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Media transformations for accounting for lagging or media effects."""

import functools
from typing import Union

import jax
import jax.numpy as jnp


@functools.partial(jax.jit, static_argnums=[0, 1])
def calculate_seasonality(
    number_periods: int,
    degrees: int,
    gamma_seasonality: Union[int, float, jnp.ndarray],
    frequency: int = 52,
) -> jnp.ndarray:
  """Calculates cyclic variation seasonality using Fourier terms.

  For detailed info check:
    https://en.wikipedia.org/wiki/Seasonality#Modeling

  Args:
    number_periods: Number of seasonal periods in the data. Eg. for 1 year of
      seasonal data it will be 52, for 3 years of the same kind 156.
    degrees: Number of degrees to use. Must be greater or equal than 1.
    gamma_seasonality: Factor to multiply to each degree calculation. Shape must
      be aligned with the number of degrees.
    frequency: Frecuency of the seasonality be in computed. By default is 52 for
      weekly data (52 weeks in a year).

  Returns:
    An array with the seasonality values.
  """

  seasonality_range = jnp.expand_dims(a=jnp.arange(number_periods), axis=-1)
  degrees_range = jnp.arange(degrees)
  inner_value = seasonality_range * 2 * jnp.pi * degrees_range / frequency
  season_matrix_sin = jnp.sin(inner_value)
  season_matrix_cos = jnp.cos(inner_value)
  season_matrix = jnp.concatenate([
      jnp.expand_dims(a=season_matrix_sin, axis=-1),
      jnp.expand_dims(a=season_matrix_cos, axis=-1)
  ],
                                  axis=-1)
  return (season_matrix * gamma_seasonality).sum(axis=2).sum(axis=1)


@jax.jit
def adstock(data: jnp.ndarray,
            lag_weight: float = .9,
            normalise: bool = True) -> jnp.ndarray:
  """Calculates the adstock value of a given array.

  To learn more about advertising lag:
  https://en.wikipedia.org/wiki/Advertising_adstock

  Args:
    data: Input array.
    lag_weight: lag_weight effect of the adstock function. Default is 0.9.
    normalise: Whether to normalise the output value. This normalization will
      divide the output values by (1 / (1 - lag_weight)).

  Returns:
    The adstock output of the input array.
  """

  def adstock_internal(prev_adstock: jnp.ndarray,
                       data: jnp.ndarray,
                       lag_weight: float = lag_weight) -> jnp.ndarray:
    adstock_value = prev_adstock * lag_weight + data
    return adstock_value, adstock_value

  _, adstock_values = jax.lax.scan(
      f=adstock_internal, init=data[0, ...], xs=data[1:, ...])
  adstock_values = jnp.concatenate([jnp.array([data[0, ...]]), adstock_values])
  return jax.lax.cond(
      normalise,
      lambda adstock_values: adstock_values / (1. / (1 - lag_weight)),
      lambda adstock_values: adstock_values,
      operand=adstock_values)


@jax.jit
def hill(data: jnp.ndarray, half_max_effective_concentration: jnp.ndarray,
         slope: jnp.ndarray) -> jnp.ndarray:
  """Calculates the hill function for a given array of values.

  Refer to the following link for detailed information on this equation:
    https://en.wikipedia.org/wiki/Hill_equation_(biochemistry)

  Args:
    data: Input data.
    half_max_effective_concentration: ec50 value for the hill function.
    slope: Slope of the hill function.

  Returns:
    The hill values for the respective input data.
  """
  save_transform = apply_exponent_safe(
      data=data / half_max_effective_concentration, exponent=-slope)
  return jnp.where(save_transform == 0, x=0, y=1. / (1 + save_transform))


@functools.partial(jax.vmap, in_axes=(1, 1, None), out_axes=1)
def _carryover_convolve(data: jnp.ndarray,
                        weights: jnp.ndarray,
                        number_lags: int) -> jnp.ndarray:
  """Applies the convolution between the data and the weights for the carryover.

  Args:
    data: Input data.
    weights: Window weights for the carryover.
    number_lags: Number of lags the window has.

  Returns:
    The result values from convolving the data and the weights with padding.
  """
  window = jnp.concatenate([jnp.zeros(number_lags - 1), weights])
  return jax.scipy.signal.convolve(data, window, mode="same") / weights.sum()


@functools.partial(jax.jit, static_argnames=("number_lags",))
def carryover(data: jnp.ndarray,
              ad_effect_retention_rate: jnp.ndarray,
              peak_effect_delay: jnp.ndarray,
              number_lags: int = 13) -> jnp.ndarray:
  """Calculates media carryover.

  More details about this function can be found in:
  https://static.googleusercontent.com/media/research.google.com/en//pubs/archive/46001.pdf

  Args:
    data: Input data. It is expected that data has either 2 dimensions for
      national models and 3 for geo models.
    ad_effect_retention_rate: Retention rate of the advertisement effect.
      Default is 0.5.
    peak_effect_delay: Delay of the peak effect in the carryover function.
      Default is 1.
    number_lags: Number of lags to include in the carryover calculation. Default
      is 13.

  Returns:
    The carryover values for the given data with the given parameters.
  """
  lags_arange = jnp.expand_dims(jnp.arange(number_lags, dtype=jnp.float32),
                                axis=-1)
  convolve_func = _carryover_convolve
  if data.ndim == 3:
    # Since _carryover_convolve is already vmaped in the decorator we only need
    # to vmap it once here to handle the geo level data. We keep the windows bi
    # dimensional also for three dims data and vmap over only the extra data
    # dimension.
    convolve_func = jax.vmap(
        fun=_carryover_convolve, in_axes=(2, None, None), out_axes=2)
  weights = ad_effect_retention_rate**((lags_arange - peak_effect_delay)**2)
  return convolve_func(data, weights, number_lags)


def apply_exponent_safe(
    data: jnp.ndarray,
    exponent: jnp.ndarray
    ) -> jnp.ndarray:
  """Applies an exponent to given data in a gradient safe way.

  More info on the double jnp.where can be found:
  https://github.com/tensorflow/probability/blob/main/discussion/where-nan.pdf

  Args:
    data: Input data to use.
    exponent: Exponent required for the operations.

  Returns:
    The result of the exponent operation with the inputs provided.
  """
  exponent_safe = jnp.where(condition=(data == 0), x=1, y=data) ** exponent
  return jnp.where(condition=(data == 0), x=0, y=exponent_safe)



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for preprocessing dataset for training LightweightMMM."""

from typing import Callable, Optional, Union

import jax.numpy as jnp

from sklearn import base


class NotFittedScalerError(Exception):
  pass


class CustomScaler(base.TransformerMixin):
  """Class to scale your data based on multiplications and divisions.

  This scaler can be used in two fashions for both the multiplication and
  division operation.
  - By specifying a value to use for the scaling operation.
  - By specifying an operation used at column level to calculate the value
    for the actual scaling operation.

  Eg. if one wants to scale the dataset by multiply by 100 you can directly
  pass multiply_by=100. Value can also be an array with as many values
  as column has the data being scaled. But if you want to multiply by the mean
  value of each column, then you can pass multiply_operation=jnp.mean (or any
  other operation desired).

  Operation parameters have the upper hand in the cases where both values and
  operations are passed, values will be ignored in this case.

  Scaler must be fit first in order to call the transform method.

  Attributes.
    divide_operation: Operation to apply over axis 0 of the fitting data to
      obtain the value that will be used for division during scaling.
    divide_by: Numbers(s) by which to divide data in the scaling process. Since
      the scaler is applied to axis 0 of the data, the shape of divide_by must
      be consistent with division into the data. For example, if data.shape =
      (100, 3, 5) then divide_by.shape can be (3, 5) or (5,) or a number. If
      divide_operation is given, this divide_by value will be ignored.
    multiply_operation: Operation to apply over axis 0 of the fitting data to
      obtain the value that will be used for multiplication during scaling.
    multiply_by: Numbers(s) by which to multiply data in the scaling process.
      Since the scaler is applied to axis 0 of the data, the shape of
      multiply_by must be consistent with multiplication into the data. For
      example, if data.shape = (100, 3, 5) then multiply_by.shape can be (3, 5)
      or (5,) or a number. If multiply_operation is given, this multiply_by
      value will be ignored.
  """

  def __init__(
      self,
      divide_operation: Optional[Callable[[jnp.ndarray], jnp.float32]] = None,
      divide_by: Optional[Union[float, int, jnp.ndarray]] = 1,
      multiply_operation: Optional[Callable[[jnp.ndarray], jnp.float32]] = None,
      multiply_by: Optional[Union[float, int, jnp.ndarray]] = 1.) -> None:
    """Constructor for the CustomScaler class."""
    if all([
        divide_by is None, divide_operation is None, multiply_by is None,
        multiply_operation is None
    ]):
      raise ValueError("No values for transformations were provided and this "
                       "scaler will fail. Please instantiate a valid one")

    if divide_operation is None and divide_by is None:
      raise ValueError(
          "Either a division operation or value needs to be passed. If "
          "you dont want to use a division to scale your data just "
          "pass divide_by=1.")
    elif divide_operation is not None:
      self.divide_operation = divide_operation
    else:
      self.divide_by = divide_by

    if multiply_operation is None and multiply_by is None:
      raise ValueError(
          "Either a multiplication operation or value needs to be passed. If "
          "you dont want to use a multiplication to scale your data just "
          "pass multiply_by=1.")
    elif multiply_operation is not None:
      self.multiply_operation = multiply_operation
    else:
      self.multiply_by = multiply_by

  def fit(self, data: jnp.ndarray) -> None:
    """Figures out values for transformations based on the specified operations.

    Args:
      data: Input dataset to use for fitting.
    """
    if hasattr(self, "divide_operation"):
      self.divide_by = jnp.apply_along_axis(
          func1d=self.divide_operation, axis=0, arr=data)
    elif isinstance(self.divide_by, int) or isinstance(self.divide_by, float):
      self.divide_by = self.divide_by * jnp.ones(data.shape[1:])
    if hasattr(self, "multiply_operation"):
      self.multiply_by = jnp.apply_along_axis(
          func1d=self.multiply_operation, axis=0, arr=data)
    elif isinstance(self.multiply_by, int) or isinstance(
        self.multiply_by, float):
      self.multiply_by = self.multiply_by * jnp.ones(data.shape[1:])

  def transform(self, data: jnp.ndarray) -> jnp.ndarray:
    """Applies transformation based on fitted values.

    It can only be called if scaler was fit first.

    Args:
      data: Input dataset to transform.

    Returns:
      Transformed array.
    """
    if not hasattr(self, "divide_by") or not hasattr(self, "multiply_by"):
      raise NotFittedScalerError(
          "transform is called without fit being called previously. Please "
          "fit scaler first.")
    return self.multiply_by * data / self.divide_by

  def fit_transform(self, data: jnp.ndarray) -> jnp.ndarray:
    """Fits the values and applies transformation to the input data.

    Args:
      data: Input dataset.

    Returns:
      Transformed array.
    """
    self.fit(data)
    return self.transform(data)

  def inverse_transform(self, data: jnp.ndarray) -> jnp.ndarray:
    """Runs inverse transformation to get original values.

    Args:
      data: Input dataset.

    Returns:
      Dataset with the inverse transformation applied.
    """
    return self.divide_by * data / self.multiply_by



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for utils."""

import os

from absl.testing import absltest
from absl.testing import parameterized
import jax.numpy as jnp
import numpy as np
import pandas as pd

from lightweight_mmm import lightweight_mmm
from lightweight_mmm import utils

_MEDIA_DATAFRAME = pd.DataFrame(
    data=[["2020-01-01", "geo1", 10, 2, 3, 2, 1, 1],
          ["2020-01-01", "geo2", 40, 6, 2, 2, 1, 0],
          ["2020-01-01", "geo3", 33, 7, 5, 2, 1, 0],
          ["2020-01-08", "geo1", 21, 1, 7, 2, 1, 1],
          ["2020-01-08", "geo2", 27, 3, 9, 2, 1, 1],
          ["2020-01-08", "geo3", 20, 5, 3, 2, 1, 1]],
    columns=[
        "date", "geo", "kpi", "channel1_imp", "channel2_imp",
        "channel1_cost", "channel2_cost", "promo_1"
    ])


class UtilsTest(parameterized.TestCase):

  def test_save_model_file_is_correctly_saved(self):
    media = jnp.ones((20, 2), dtype=jnp.float32)
    extra_features = jnp.arange(20).reshape((20, 1))
    costs = jnp.arange(1, 3)
    target = jnp.arange(1, 21)
    mmm_object = lightweight_mmm.LightweightMMM()
    mmm_object.fit(
        media=media,
        extra_features=extra_features,
        media_prior=costs,
        target=target,
        number_warmup=10,
        number_samples=100,
        number_chains=1)

    file_path = os.path.join(self.create_tempdir().full_path, "model.pkl")
    utils.save_model(media_mix_model=mmm_object,
                     file_path=file_path)

    self.assertTrue(os.path.exists(file_path))

  def test_load_model_with_all_attributes(self):
    media = jnp.ones((20, 2), dtype=jnp.float32)
    extra_features = jnp.arange(20).reshape((20, 1))
    costs = jnp.arange(1, 3)
    target = jnp.arange(1, 21)
    mmm_object = lightweight_mmm.LightweightMMM()
    mmm_object.fit(
        media=media,
        extra_features=extra_features,
        media_prior=costs,
        target=target,
        number_warmup=10,
        number_samples=100,
        number_chains=1)
    file_path = os.path.join(self.create_tempdir().full_path, "model.pkl")
    utils.save_model(media_mix_model=mmm_object,
                     file_path=file_path)

    loaded_mmm = utils.load_model(file_path)

    self.assertEqual(mmm_object, loaded_mmm)

  @parameterized.named_parameters([
      dict(
          testcase_name="shape_100_3_3",
          data_size=100,
          n_media_channels=3,
          n_extra_features=3),
      dict(
          testcase_name="shape_200_8_1",
          data_size=200,
          n_media_channels=8,
          n_extra_features=1),
      dict(
          testcase_name="shape_300_2_2",
          data_size=300,
          n_media_channels=2,
          n_extra_features=2),
      dict(
          testcase_name="shape_400_4_10",
          data_size=400,
          n_media_channels=4,
          n_extra_features=10)
  ])
  def test_simulate_dummy_data_produces_correct_shape(self,
                                                      data_size,
                                                      n_media_channels,
                                                      n_extra_features):
    media_data, extra_features, target, costs = utils.simulate_dummy_data(
        data_size=data_size,
        n_media_channels=n_media_channels,
        n_extra_features=n_extra_features)

    self.assertEqual(media_data.shape, (data_size, n_media_channels))
    self.assertEqual(extra_features.shape, (data_size, n_extra_features))
    self.assertEqual(target.shape, (data_size,))
    self.assertLen(costs, n_media_channels)

  @parameterized.named_parameters([
      dict(
          testcase_name="array_shape_without_cost_feature",
          media_features=["channel1_imp", "channel2_imp"],
          extra_features=["promo_1"],
          geo_feature="geo",
          date_feature="date",
          target="kpi",
          cost_features=None),
      dict(
          testcase_name="array_shape_with_cost_feature",
          media_features=["channel1_imp", "channel2_imp"],
          extra_features=["promo_1"],
          geo_feature="geo",
          date_feature="date",
          target="kpi",
          cost_features=["channel1_cost", "channel2_cost"])
  ])
  def test_dataframe_to_jax_produce_correct_shape_with_or_without_cost(
      self, media_features, extra_features, geo_feature, date_feature, target,
      cost_features):
    media_data, extra_features_data, target_data, costs_data = utils.dataframe_to_jax(
        dataframe=_MEDIA_DATAFRAME,
        media_features=media_features,
        extra_features=extra_features,
        geo_feature=geo_feature,
        date_feature=date_feature,
        target=target,
        cost_features=cost_features
        )
    n_weeks = _MEDIA_DATAFRAME[date_feature].nunique()
    n_geos = _MEDIA_DATAFRAME[geo_feature].nunique()
    n_media_channels = len(media_features)
    n_extra_features = len(extra_features)
    self.assertEqual(media_data.shape, (n_weeks, n_media_channels, n_geos))
    self.assertEqual(extra_features_data.shape,
                     (n_weeks, n_extra_features, n_geos))
    self.assertEqual(target_data.shape, (n_weeks, n_geos))
    self.assertLen(costs_data, n_media_channels)

  @parameterized.named_parameters([
      dict(
          testcase_name="array_value_without_cost_feature",
          media_features=["channel1_imp", "channel2_imp"],
          extra_features=["promo_1"],
          geo_feature="geo",
          date_feature="date",
          target="kpi",
          cost_features=None,
          test_geo="geo1"),
      dict(
          testcase_name="array_value_with_cost_feature",
          media_features=["channel1_imp", "channel2_imp"],
          extra_features=["promo_1"],
          geo_feature="geo",
          date_feature="date",
          target="kpi",
          cost_features=["channel1_cost", "channel2_cost"],
          test_geo="geo1")
  ])
  def test_dataframe_to_jax_produce_correct_value_with_or_without_cost(
      self, media_features, extra_features, geo_feature, date_feature, target,
      cost_features, test_geo):
    media_data, extra_features_data, target_data, costs_data = utils.dataframe_to_jax(
        dataframe=_MEDIA_DATAFRAME,
        media_features=media_features,
        extra_features=extra_features,
        geo_feature=geo_feature,
        date_feature=date_feature,
        target=target,
        cost_features=cost_features)
    n_weeks = _MEDIA_DATAFRAME[date_feature].nunique()
    geo1_media = _MEDIA_DATAFRAME.loc[_MEDIA_DATAFRAME[geo_feature] == test_geo,
                                      media_features].values
    geo1_extra_features = _MEDIA_DATAFRAME.loc[_MEDIA_DATAFRAME[geo_feature] ==
                                               test_geo, extra_features].values
    geo1_target = _MEDIA_DATAFRAME.loc[_MEDIA_DATAFRAME[geo_feature] ==
                                       test_geo, [target]].values
    geo1_target = geo1_target.reshape(n_weeks,)
    if cost_features is None:
      cost_by_channel = _MEDIA_DATAFRAME.loc[:, media_features].sum().values
    else:
      cost_by_channel = _MEDIA_DATAFRAME.loc[:, cost_features].sum().values

    np.testing.assert_array_equal(media_data[:, :, 0], geo1_media)
    np.testing.assert_array_equal(extra_features_data[:, :, 0],
                                  geo1_extra_features)
    np.testing.assert_array_equal(target_data[:, 0],
                                  geo1_target.reshape(n_weeks,))
    np.testing.assert_array_equal(costs_data, cost_by_channel)

  @parameterized.named_parameters([
      dict(
          testcase_name="shape_0_3_3",
          data_size=0,
          n_media_channels=3,
          n_extra_features=3),
      dict(
          testcase_name="shape_200_-1_1",
          data_size=200,
          n_media_channels=-1,
          n_extra_features=1),
      dict(
          testcase_name="shape_300_2_-2",
          data_size=300,
          n_media_channels=2,
          n_extra_features=-2),
      dict(
          testcase_name="shape_-400_-4_-10",
          data_size=-400,
          n_media_channels=-4,
          n_extra_features=-10)
  ])
  def test_simulate_dummy_data_with_zero_or_neg_parameter_raises_value_error(
      self, data_size, n_media_channels, n_extra_features):

    with self.assertRaises(ValueError):
      utils.simulate_dummy_data(
          data_size=data_size,
          n_media_channels=n_media_channels,
          n_extra_features=n_extra_features)

  def test_simulate_geo_data_has_right_shape(self):
    data_size = 100
    geos = 3
    media_data, _, target, _ = utils.simulate_dummy_data(
        data_size, 2, 2, geos=geos)
    self.assertEqual(target.shape, (data_size, geos))
    self.assertEqual(media_data.shape, (data_size, 2, geos))

  def test_halfnormal_mean_and_scale(self):
    mean = 1.
    scale = utils.get_halfnormal_scale_from_mean(mean)
    new_mean = utils.get_halfnormal_mean_from_scale(scale)
    self.assertEqual(scale, mean * np.sqrt(np.pi) / np.sqrt(2))
    self.assertEqual(mean, new_mean)

  def test_beta_params_match(self):
    a, b = 2., 3.
    # Expected mean is 2 / 5.
    mu = a / (a + b)
    sigma = np.sqrt(a * b / ((a + b) ** 2 * (a + b + 1)))
    ahat, bhat = utils.get_beta_params_from_mu_sigma(mu, sigma)
    self.assertAlmostEqual(ahat / (ahat + bhat), 2 / 5)

  def test_prior_posterior_distance_discrete(self):
    p = jnp.array([0] * 2 + [1] * 3)
    q = jnp.array([0] * 3 + [1] * 2 + [2] * 1)
    ks = utils.distance_pior_posterior(p, q, method="KS", discrete=True)
    js = utils.distance_pior_posterior(p, q, method="JS", discrete=True)
    hell = utils.distance_pior_posterior(
        p, q, method="Hellinger", discrete=True)
    mindist = utils.distance_pior_posterior(p, q, method="min", discrete=True)
    print(ks, js, hell, mindist)
    self.assertAlmostEqual(ks, 1 / 6)
    self.assertAlmostEqual(js, 0.283, 3)
    self.assertAlmostEqual(hell, 0.325, 3)
    self.assertAlmostEqual(mindist, 0.267, 3)

  def test_prior_posterior_distance_continuous(self):
    p = jnp.array([0] * 2 + [.5] * 3 + [1] * 2)
    q = jnp.array([0] * 2 + [.5] * 4 + [1] * 2 + [1.5] * 3)
    ks = utils.distance_pior_posterior(p, q, method="KS", discrete=False)
    js = utils.distance_pior_posterior(p, q, method="JS", discrete=False)
    hell = utils.distance_pior_posterior(
        p, q, method="Hellinger", discrete=False)
    mindist = utils.distance_pior_posterior(p, q, method="min", discrete=False)
    print(ks, js, hell, mindist)
    self.assertAlmostEqual(ks, 0.2727, 4)
    self.assertAlmostEqual(js, 0.034, 3)
    self.assertAlmostEqual(hell, 0.034, 3)
    self.assertAlmostEqual(mindist, 0.041, 3)

  def test_outlier_interpolation_straight_line(self):
    x = np.arange(10) * 1.
    x[3:5] += 10
    x = jnp.array(x)
    outlier_idx = jnp.array([3, 4])
    new_x = utils.interpolate_outliers(x, outlier_idx)
    self.assertTrue(all(np.equal(new_x[outlier_idx], [3, 4])))


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for media_transforms."""

from absl.testing import absltest
from absl.testing import parameterized
import jax
import jax.numpy as jnp
import numpy as np

from lightweight_mmm import media_transforms


class MediaTransformsTest(parameterized.TestCase):

  @parameterized.named_parameters([
      dict(
          testcase_name="2d_four_channels",
          data=np.ones((100, 4)),
          ad_effect_retention_rate=np.array([0.9, 0.8, 0.7, 1]),
          peak_effect_delay=np.array([0.9, 0.8, 0.7, 1]),
          number_lags=5),
      dict(
          testcase_name="2d_one_channel",
          data=np.ones((300, 1)),
          ad_effect_retention_rate=np.array([0.2]),
          peak_effect_delay=np.array([1]),
          number_lags=10),
      dict(
          testcase_name="3d_10channels_10geos",
          data=np.ones((100, 10, 10)),
          ad_effect_retention_rate=np.ones(10),
          peak_effect_delay=np.ones(10),
          number_lags=13),
      dict(
          testcase_name="3d_10channels_8geos",
          data=np.ones((100, 10, 8)),
          ad_effect_retention_rate=np.ones(10),
          peak_effect_delay=np.ones(10),
          number_lags=13),
  ])
  def test_carryover_produces_correct_shape(self, data,
                                            ad_effect_retention_rate,
                                            peak_effect_delay, number_lags):

    generated_output = media_transforms.carryover(data,
                                                  ad_effect_retention_rate,
                                                  peak_effect_delay,
                                                  number_lags)
    self.assertEqual(generated_output.shape, data.shape)

  @parameterized.named_parameters([
      dict(
          testcase_name="2d_three_channels",
          data=np.ones((100, 3)),
          half_max_effective_concentration=np.array([0.9, 0.8, 0.7]),
          slope=np.array([2, 2, 1])),
      dict(
          testcase_name="2d_one_channels",
          data=np.ones((100, 1)),
          half_max_effective_concentration=np.array([0.9]),
          slope=np.array([5])),
      dict(
          testcase_name="3d_10channels_5geos",
          data=np.ones((100, 10, 5)),
          half_max_effective_concentration=np.expand_dims(np.ones(10), axis=-1),
          slope=np.expand_dims(np.ones(10), axis=-1)),
      dict(
          testcase_name="3d_8channels_10geos",
          data=np.ones((100, 8, 10)),
          half_max_effective_concentration=np.expand_dims(np.ones(8), axis=-1),
          slope=np.expand_dims(np.ones(8), axis=-1)),
  ])
  def test_hill_produces_correct_shape(self, data,
                                       half_max_effective_concentration, slope):
    generated_output = media_transforms.hill(
        data=data,
        half_max_effective_concentration=half_max_effective_concentration,
        slope=slope)

    self.assertEqual(generated_output.shape, data.shape)

  @parameterized.named_parameters([
      dict(
          testcase_name="2d_five_channels",
          data=np.ones((100, 5)),
          lag_weight=np.array([0.2, 0.3, 0.8, 0.2, 0.1]),
          normalise=True),
      dict(
          testcase_name="2d_one_channels",
          data=np.ones((100, 1)),
          lag_weight=np.array([0.4]),
          normalise=False),
      dict(
          testcase_name="3d_10channels_5geos",
          data=np.ones((100, 10, 5)),
          lag_weight=np.expand_dims(np.ones(10), axis=-1),
          normalise=True),
      dict(
          testcase_name="3d_8channels_10geos",
          data=np.ones((100, 8, 10)),
          lag_weight=np.expand_dims(np.ones(8), axis=-1),
          normalise=True),
  ])
  def test_adstock_produces_correct_shape(self, data, lag_weight, normalise):
    generated_output = media_transforms.adstock(
        data=data, lag_weight=lag_weight, normalise=normalise)

    self.assertEqual(generated_output.shape, data.shape)

  def test_apply_exponent_safe_produces_correct_shape(self):
    data = jnp.arange(50).reshape((10, 5))
    exponent = jnp.full(5, 0.5)

    output = media_transforms.apply_exponent_safe(data=data, exponent=exponent)

    np.testing.assert_array_equal(x=output, y=data**exponent)

  def test_apply_exponent_safe_produces_same_exponent_results(self):
    data = jnp.ones((10, 5))
    exponent = jnp.full(5, 0.5)

    output = media_transforms.apply_exponent_safe(data=data, exponent=exponent)

    self.assertEqual(output.shape, data.shape)

  def test_apply_exponent_safe_produces_non_nan_or_inf_grads(self):
    def f_safe(data, exponent):
      x = media_transforms.apply_exponent_safe(data=data, exponent=exponent)
      return x.sum()
    data = jnp.ones((10, 5))
    data = data.at[0, 0].set(0.)
    exponent = jnp.full(5, 0.5)

    grads = jax.grad(f_safe)(data, exponent)

    self.assertFalse(np.isnan(grads).any())
    self.assertFalse(np.isinf(grads).any())

  def test_adstock_zeros_stay_zeros(self):
    data = jnp.zeros((10, 5))
    lag_weight = jnp.full(5, 0.5)

    generated_output = media_transforms.adstock(
        data=data, lag_weight=lag_weight)

    np.testing.assert_array_equal(x=generated_output, y=data)

  def test_hill_zeros_stay_zeros(self):
    data = jnp.zeros((10, 5))
    half_max_effective_concentration = jnp.full(5, 0.5)
    slope = jnp.full(5, 0.5)

    generated_output = media_transforms.hill(
        data=data,
        half_max_effective_concentration=half_max_effective_concentration,
        slope=slope)

    np.testing.assert_array_equal(x=generated_output, y=data)

  def test_carryover_zeros_stay_zeros(self):
    data = jnp.zeros((10, 5))
    ad_effect_retention_rate = jnp.full(5, 0.5)
    peak_effect_delay = jnp.full(5, 0.5)

    generated_output = media_transforms.carryover(
        data=data,
        ad_effect_retention_rate=ad_effect_retention_rate,
        peak_effect_delay=peak_effect_delay)

    np.testing.assert_array_equal(x=generated_output, y=data)


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for plot."""

from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
import jax.numpy as jnp
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import numpyro.distributions as dist

from lightweight_mmm import lightweight_mmm
from lightweight_mmm import plot
from lightweight_mmm import preprocessing


class PlotTest(parameterized.TestCase):

  @classmethod
  def setUpClass(cls):
    super(PlotTest, cls).setUpClass()
    cls.national_mmm = lightweight_mmm.LightweightMMM()
    cls.national_mmm.fit(
        media=jnp.ones((50, 5)),
        target=jnp.ones(50),
        media_prior=jnp.ones(5) * 50,
        number_warmup=2,
        number_samples=2,
        number_chains=1)
    cls.geo_mmm = lightweight_mmm.LightweightMMM()
    cls.geo_mmm.fit(
        media=jnp.ones((50, 5, 3)),
        target=jnp.ones((50, 3)),
        media_prior=jnp.ones(5) * 50,
        number_warmup=2,
        number_samples=2,
        number_chains=1)
    cls.not_fitted_mmm = lightweight_mmm.LightweightMMM()

  def setUp(self):
    super().setUp()
    self.mock_ax_scatter = self.enter_context(
        mock.patch.object(plot.plt.Axes, "scatter", autospec=True))
    self.mock_sns_lineplot = self.enter_context(
        mock.patch.object(plot.sns, "lineplot", autospec=True))
    self.mock_plt_plot = self.enter_context(
        mock.patch.object(plot.plt.Axes, "plot", autospec=True))
    self.mock_plt_barplot = self.enter_context(
        mock.patch.object(plot.plt.Axes, "bar", autospec=True))
    self.mock_pd_area_plot = self.enter_context(
        mock.patch.object(plot.pd.DataFrame.plot, "area", autospec=True))
    self.mock_sns_kdeplot = self.enter_context(
        mock.patch.object(plot.sns, "kdeplot", autospec=True))
    self.mock_plt_ax_legend = self.enter_context(
        mock.patch.object(plot.plt.Axes, "legend", autospec=True))

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          media_mix_model="national_mmm",
          expected_calls=1),
      dict(testcase_name="geo", media_mix_model="geo_mmm", expected_calls=2)
  ])

  def test_plot_model_fit_plot_called_with_scaler(self, media_mix_model,
                                                  expected_calls):
    target_scaler = preprocessing.CustomScaler(divide_operation=jnp.mean)
    target_scaler.fit(jnp.ones(1))
    mmm = getattr(self, media_mix_model)

    plot.plot_model_fit(media_mix_model=mmm, target_scaler=target_scaler)

    self.assertTrue(self.mock_plt_plot.call_count, expected_calls)

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          media_mix_model="national_mmm",
          expected_calls=1),
      dict(
          testcase_name="geo",
          media_mix_model="geo_mmm",
          expected_calls=2)
  ])
  def test_plot_model_fit_plot_called_without_scaler(
      self, media_mix_model, expected_calls):
    mmm = getattr(self, media_mix_model)

    plot.plot_model_fit(media_mix_model=mmm)

    self.assertTrue(self.mock_plt_plot.call_count, expected_calls)

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          media_mix_model="national_mmm"),
      dict(
          testcase_name="geo",
          media_mix_model="geo_mmm")
  ])
  def test_plot_response_curves_plots_n_times_with_correct_params(
      self, media_mix_model):
    mmm = getattr(self, media_mix_model)
    n_geos = (mmm.media.shape[-1] if mmm.media.ndim == 3
              else 1)
    plot.plot_response_curves(media_mix_model=mmm)

    _, call_kwargs = self.mock_sns_lineplot.call_args_list[0]
    # n channels times 2 charts.
    self.assertEqual(self.mock_sns_lineplot.call_count,
                     2 * mmm.n_media_channels)
    self.assertEqual(jnp.round(a=call_kwargs["x"].max(), decimals=4),
                     jnp.round(a=1.2 * n_geos, decimals=4))

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          media_mix_model="national_mmm"),
      dict(
          testcase_name="geo",
          media_mix_model="geo_mmm")
  ])
  def test_plot_response_curves_with_prices_plots_n_times_with_correct_params(
      self, media_mix_model="geo_mmm"):
    mmm = getattr(self, media_mix_model)
    n_channels = mmm.n_media_channels
    n_geos = mmm.media.shape[-1] if mmm.media.ndim == 3 else 1
    prices = jnp.array([1., 0.8, 2., 3., 1.])
    expected_maxes = jnp.repeat(jnp.array([1.2, 0.96, 2.4, 3.6, 1.2]), 2)

    plot.plot_response_curves(media_mix_model=mmm, prices=prices)

    calls_list = self.mock_sns_lineplot.call_args_list
    self.assertEqual(self.mock_sns_lineplot.call_count, n_channels * 2)
    for (_, call_kwargs), expected_max in zip(calls_list, expected_maxes):
      self.assertAlmostEqual(
          jnp.round(a=call_kwargs["x"].max().item(), decimals=4).item(),
          jnp.round(a=expected_max, decimals=4).item() * n_geos,
          places=4)

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          media_mix_model="national_mmm"),
      dict(
          testcase_name="geo",
          media_mix_model="geo_mmm")
  ])
  def test_plot_response_curves_produces_y_axis_starting_at_zero(
      self, media_mix_model):
    mmm = getattr(self, media_mix_model)

    plot.plot_response_curves(media_mix_model=mmm)

    calls_list = self.mock_sns_lineplot.call_args_list
    for _, call_kwargs in calls_list[:3]:
      self.assertLessEqual(call_kwargs["y"].min().item(), 0.1)
      self.assertGreaterEqual(call_kwargs["y"].min().item(), -0.1)

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          media_mix_model="national_mmm"),
      dict(
          testcase_name="geo",
          media_mix_model="geo_mmm")
  ])
  def test_plot_response_curves_scales_with_media_scaler(
      self, media_mix_model):
    media_scaler = preprocessing.CustomScaler(divide_operation=jnp.mean)
    factor = 5
    media_scaler.fit(jnp.ones(1) * factor)
    expected_maxes = jnp.repeat(
        jnp.repeat(jnp.array([1.2]), repeats=5),
        repeats=2)
    mmm = getattr(self, media_mix_model)

    plot.plot_response_curves(media_mix_model=mmm,
                              media_scaler=media_scaler)

    calls_list = self.mock_plt_plot.call_args_list
    for (_, call_kwargs), expected_max in zip(calls_list, expected_maxes):
      self.assertAlmostEqual(call_kwargs["x"].max().item(),
                             expected_max * factor,
                             places=4)

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          media_mix_model="national_mmm"),
      dict(
          testcase_name="geo",
          media_mix_model="geo_mmm")
  ])
  def test_plot_response_curves_scales_with_target_scaler(
      self, media_mix_model):
    target_scaler = preprocessing.CustomScaler(divide_operation=jnp.mean)
    factor = 5
    target_scaler.fit(jnp.ones(1) * factor)
    mmm = getattr(self, media_mix_model)

    plot.plot_response_curves(media_mix_model=mmm,
                              target_scaler=target_scaler)

    calls_list = self.mock_plt_plot.call_args_list
    for _, call_kwargs in calls_list:
      self.assertAlmostEqual(call_kwargs["y"].max().item(),
                             1 * factor,
                             places=4)

  def test_perfect_correlation_returns_correct_output(self):
    x = jnp.arange(100)
    y = jnp.arange(100, 200)

    idx, maxcorr = plot.plot_cross_correlate(x, y)

    self.assertEqual(idx, 0)
    self.assertEqual(maxcorr, 1)

  def test_var_cost_plot_called_with_correct_kwargs(self):
    media = jnp.arange(10).reshape((5, 2))
    costs = [1, 2]
    names = ["a", "b"]
    std = jnp.repeat(2.82842712, 2)
    means = jnp.array([4, 5])
    expected_coef_of_variation = std / means

    _ = plot.plot_var_cost(media, costs, names)
    _, call_kwargs = self.mock_ax_scatter.call_args_list[0]

    np.testing.assert_array_almost_equal(call_kwargs["x"], costs)
    np.testing.assert_array_almost_equal(call_kwargs["y"],
                                         expected_coef_of_variation)

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          media_mix_model="national_mmm"),
      dict(
          testcase_name="geo",
          media_mix_model="geo_mmm")
  ])
  def test_plot_media_channel_posteriors_plots_right_number_subplots(
      self, media_mix_model):
    mmm = getattr(self, media_mix_model)
    n_channels = mmm.n_media_channels
    n_geos = mmm.media.shape[-1] if mmm.media.ndim == 3 else 1

    fig = plot.plot_media_channel_posteriors(media_mix_model=mmm)

    self.assertLen(fig.get_axes(), n_channels * n_geos)

  def test_unequal_length_ground_truth_and_predictions_raises_error(self):
    prediction = jnp.arange(10).reshape((5, 2))
    ground_truth = jnp.array([1, 2, 3])
    with self.assertRaises(ValueError):
      plot.plot_out_of_sample_model_fit(prediction, ground_truth)

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          media_mix_model="national_mmm",
          expected_calls=3),
      dict(testcase_name="geo", media_mix_model="geo_mmm", expected_calls=3)
  ])
  def test_plot_pre_post_budget_allocation_comparison_n_times_with_correct_params(
      self, media_mix_model, expected_calls):
    mmm = getattr(self, media_mix_model)
    kpi_with_optim = -503
    kpi_without_optim = -479
    optimal_buget_allocation = jnp.array([118, 278, 100, 100, 100])
    previous_budget_allocation = jnp.array([199, 197, 100, 100, 100])

    plot.plot_pre_post_budget_allocation_comparison(
        media_mix_model=mmm,
        kpi_with_optim=kpi_with_optim,
        kpi_without_optim=kpi_without_optim,
        optimal_buget_allocation=optimal_buget_allocation,
        previous_budget_allocation=previous_budget_allocation)

    self.assertEqual(self.mock_plt_barplot.call_count, expected_calls)
    call_list = self.mock_plt_barplot.call_args_list

    np.testing.assert_array_almost_equal(call_list[0][0][-1],
                                         previous_budget_allocation)
    np.testing.assert_array_almost_equal(call_list[1][0][-1],
                                         optimal_buget_allocation)
    np.testing.assert_array_almost_equal(
        call_list[2][0][-1], [kpi_without_optim * -1, kpi_with_optim * -1])

  @parameterized.named_parameters([
      dict(testcase_name="national and geo", media_mix_model="not_fitted_mmm")
  ])
  def test_plot_pre_post_budget_allocation_comparison_raise_notfittedmodelerror(
      self, media_mix_model):
    mmm = getattr(self, media_mix_model)
    kpi_with_optim = -503
    kpi_without_optim = -479
    optimal_buget_allocation = jnp.array([118, 278, 100, 100, 100])
    previous_budget_allocation = jnp.array([199, 197, 100, 100, 100])
    with self.assertRaises(lightweight_mmm.NotFittedModelError):
      plot.plot_pre_post_budget_allocation_comparison(
          media_mix_model=mmm,
          kpi_with_optim=kpi_with_optim,
          kpi_without_optim=kpi_without_optim,
          optimal_buget_allocation=optimal_buget_allocation,
          previous_budget_allocation=previous_budget_allocation)

  @parameterized.named_parameters([
      dict(testcase_name="national", media_mix_model="national_mmm"),
      dict(testcase_name="geo", media_mix_model="geo_mmm")
  ])
  def test_create_media_baseline_contribution_df_contributions_add_up_avg_prediction(
      self, media_mix_model):
    mmm = getattr(self, media_mix_model)
    target_scaler = preprocessing.CustomScaler(divide_operation=jnp.mean)
    target_scaler.fit(jnp.ones(1))
    media_channels_baseline_contribution_df = plot.create_media_baseline_contribution_df(
        media_mix_model=mmm, target_scaler=target_scaler)
    contribution_pct_cols = [
        col for col in media_channels_baseline_contribution_df.columns
        if "percentage" in col
    ]
    contribution_cols = [
        col for col in media_channels_baseline_contribution_df.columns
        if "contribution" in col
    ]
    # test whether contribution percentage sums up to 1 for each period
    np.testing.assert_array_almost_equal(
        media_channels_baseline_contribution_df[contribution_pct_cols].sum(
            axis=1),
        jnp.repeat(1, media_channels_baseline_contribution_df.shape[0]))
    # test whether contribution volume sums up to avg predition for each period
    np.testing.assert_array_almost_equal(
        np.round(
            media_channels_baseline_contribution_df[contribution_cols].sum(
                axis=1), 0),
        np.round(media_channels_baseline_contribution_df["avg_prediction"], 0))

  def test_create_media_baseline_contribution_df_returns_accurate_contribution_pct(
      self):
    mmm_object = lightweight_mmm.LightweightMMM()
    mmm_object.media = jnp.ones((1, 3))
    mmm_object._total_costs = jnp.array([2., 1., 3.]) * 15
    mmm_object._target = jnp.ones((1, 1)) * 5
    mmm_object.media_names = ["channel_0", "channel_1", "channel_2"]
    mmm_object.trace = {
        "media_transformed": jnp.ones((500, 1, 3)) * jnp.arange(1, 4),
        "mu": jnp.ones((500, 1)) * 10,
        "coef_media": jnp.ones((500, 3)) * 0.5
    }
    expected_contribution_pct = jnp.array([0.05, 0.1, 0.15])

    contribution_df = plot.create_media_baseline_contribution_df(
        media_mix_model=mmm_object)
    contribution_percentage_cols = [
        "{}_percentage".format(col) for col in mmm_object.media_names
    ]
    np.testing.assert_array_almost_equal(
        expected_contribution_pct,
        contribution_df[contribution_percentage_cols].values.flatten().tolist())

  @parameterized.named_parameters([
      dict(
          testcase_name="national",
          media_mix_model="national_mmm",
          expected_calls=1),
      dict(testcase_name="geo", media_mix_model="geo_mmm", expected_calls=1)
  ])
  def test_plot_media_baseline_contribution_area_plot(self, media_mix_model,
                                                      expected_calls):
    mmm = getattr(self, media_mix_model)
    target_scaler = preprocessing.CustomScaler(divide_operation=jnp.mean)
    target_scaler.fit(jnp.ones(1))
    _ = plot.plot_media_baseline_contribution_area_plot(
        media_mix_model=mmm, target_scaler=target_scaler, legend_outside=True)
    self.assertEqual(self.mock_pd_area_plot.call_count, expected_calls)

  def test_legend_plot_media_baseline_contribution_area_plot(self):
    mmm = getattr(self, "national_mmm")
    target_scaler = preprocessing.CustomScaler(divide_operation=jnp.mean)
    target_scaler.fit(jnp.ones(1))
    _ = plot.plot_media_baseline_contribution_area_plot(
        media_mix_model=mmm, target_scaler=target_scaler, legend_outside=True)
    call_args, call_kwargs = self.mock_plt_ax_legend.call_args_list[0]
    self.assertEqual(call_kwargs["loc"], "center left")
    self.assertEqual(call_kwargs["bbox_to_anchor"], (1, 0.5))


  def test_plot_response_curves_works_twice_with_non_jnp_data(self):
    mmm_object = lightweight_mmm.LightweightMMM()
    mmm_object.fit(
        media=np.ones((50, 5)),
        target=np.ones(50),
        media_prior=np.ones(5) * 50,
        number_warmup=2,
        number_samples=2,
        number_chains=1)
    _ = plot.plot_response_curves(media_mix_model=mmm_object)
    fig = plot.plot_response_curves(media_mix_model=mmm_object)
    self.assertIsInstance(fig, matplotlib.figure.Figure)

  @parameterized.named_parameters([
      dict(testcase_name="trace_missing", missing_attribute="trace"),
      dict(
          testcase_name="weekday_seasonality_missing",
          missing_attribute="_weekday_seasonality"),
      dict(
          testcase_name="custom_priors_missing",
          missing_attribute="custom_priors"),
      dict(testcase_name="n_geos_missing", missing_attribute="n_geos"),
      dict(
          testcase_name="n_media_channels_missing",
          missing_attribute="n_media_channels"),
      dict(
          testcase_name="media_prior_missing",
          missing_attribute="_media_prior"),
  ])
  def test_prior_posterior_plot_raises_notfittedmodelerror(
      self, missing_attribute):
    mmm = getattr(self, "not_fitted_mmm")
    mmm.trace = jnp.ones((50, 5))
    mmm._weekday_seasonality = True
    mmm.custom_priors = None
    mmm.n_geos = 1
    mmm.n_media_channels = 3
    mmm._media_prior = jnp.ones(5) * 50

    with self.assertRaises(lightweight_mmm.NotFittedModelError):
      delattr(mmm, missing_attribute)
      plot.plot_prior_and_posterior(
          media_mix_model=mmm, number_of_samples_for_prior=100)

  @parameterized.named_parameters([
      dict(
          testcase_name="half_normal_prior",
          prior_distribution=dist.HalfNormal(3),
          expected_clipping_bounds=[0, None]),
      dict(
          testcase_name="normal_prior",
          prior_distribution=dist.Normal(0, 1),
          expected_clipping_bounds=None),
      dict(
          testcase_name="beta_prior",
          prior_distribution="beta distribution placeholder",
          expected_clipping_bounds=[0, 1]),
      dict(
          testcase_name="gamma_prior",
          prior_distribution=dist.Gamma(0.5),
          expected_clipping_bounds=[0, None]),
  ])
  def test_prior_posterior_plot_clipping_bounds_for_kdeplots(
      self, prior_distribution, expected_clipping_bounds):
    fig = plt.figure()
    gridspec_fig = matplotlib.gridspec.GridSpec(
        nrows=1, ncols=1, figure=fig, hspace=10)

    # dist.Beta() calls jnp.broadcast_to() upon instantiation, so this
    # distribution can't be left in the decorator or it raises a
    # RuntimeError: "Attempted call to JAX before absl.app.run() is called".
    # Thus we have to instantiate it inside the unit test instead.
    if prior_distribution == "beta distribution placeholder":
      prior_distribution = dist.Beta(0.5, 0.5)

    plot._make_prior_and_posterior_subplot_for_one_feature(
        prior_distribution=prior_distribution,
        posterior_samples=jnp.ones(50),
        subplot_title="title",
        fig=fig,
        gridspec_fig=gridspec_fig,
        i_ax=0, number_of_samples_for_prior=100)
    call_details = self.mock_sns_kdeplot.call_args_list
    called_clipping_bounds = call_details[0][1]["clip"]

    self.assertEqual(called_clipping_bounds, expected_clipping_bounds)

  @parameterized.named_parameters([
      dict(
          testcase_name="carryover_national_model",
          model_name="carryover",
          is_geo_model=False,
          expected_number_of_subplots=31),
      dict(
          testcase_name="carryover_geo_model",
          model_name="carryover",
          is_geo_model=True,
          expected_number_of_subplots=47),
      dict(
          testcase_name="adstock_national_model",
          model_name="adstock",
          is_geo_model=False,
          expected_number_of_subplots=26),
      dict(
          testcase_name="adstock_geo_model",
          model_name="adstock",
          is_geo_model=True,
          expected_number_of_subplots=42),
      dict(
          testcase_name="hill_adstock_national_model",
          model_name="hill_adstock",
          is_geo_model=False,
          expected_number_of_subplots=31),
      dict(
          testcase_name="hill_adstock_geo_model",
          model_name="hill_adstock",
          is_geo_model=True,
          expected_number_of_subplots=47),
  ])
  def test_prior_posterior_plot_makes_correct_number_of_subplots(
      self, model_name, is_geo_model, expected_number_of_subplots):

    mmm = lightweight_mmm.LightweightMMM(model_name=model_name)
    if is_geo_model:
      media = jnp.ones((50, 5, 3))
      target = jnp.ones((50, 3))
      extra_features = jnp.ones((50, 1, 3))
    else:
      media = jnp.ones((50, 5))
      target = jnp.ones(50)
      extra_features = jnp.ones((50, 1))
    mmm.fit(
        media=media,
        target=target,
        media_prior=jnp.ones(5) * 50,
        extra_features=extra_features,
        degrees_seasonality=3,
        number_warmup=2,
        number_samples=2,
        number_chains=1)

    fig = plot.plot_prior_and_posterior(
        media_mix_model=mmm, number_of_samples_for_prior=100, seed=0)

    self.assertLen(fig.get_axes(), expected_number_of_subplots)

  @parameterized.product(
      (dict(is_geo_model=True, expected_number_of_subplots=11),
       dict(is_geo_model=False, expected_number_of_subplots=7)),
      (dict(model_name="adstock",
            selected_features=["sigma", "intercept", "exponent"]),
       dict(model_name="carryover",
            selected_features=["sigma", "intercept", "exponent"]),
       dict(model_name="hill_adstock",
            selected_features=["sigma", "intercept", "slope"]))
      )
  def test_selected_features_for_prior_posterior_plot_makes_correct_number_of_subplots(
      self, model_name, selected_features, is_geo_model,
      expected_number_of_subplots):

    mmm = lightweight_mmm.LightweightMMM(model_name=model_name)
    if is_geo_model:
      media = jnp.ones((50, 5, 3))
      target = jnp.ones((50, 3))
      extra_features = jnp.ones((50, 1, 3))
    else:
      media = jnp.ones((50, 5))
      target = jnp.ones(50)
      extra_features = jnp.ones((50, 1))
    mmm.fit(
        media=media,
        target=target,
        media_prior=jnp.ones(5) * 50,
        extra_features=extra_features,
        number_warmup=2,
        number_samples=2,
        number_chains=1)

    fig = plot.plot_prior_and_posterior(
        media_mix_model=mmm,
        number_of_samples_for_prior=100,
        seed=0,
        selected_features=selected_features)

    self.assertLen(fig.get_axes(), expected_number_of_subplots)

  @parameterized.named_parameters([
      dict(
          testcase_name="national_model",
          model_name="national_mmm"),
      dict(
          testcase_name="geo_model",
          model_name="geo_mmm"),
  ])
  def test_selected_features_raises_value_error(self, model_name):
    mmm = getattr(self, model_name)

    # The expected error message here is an f-string whose value is filled in by
    # a set that contains a single element. The message should look exactly as
    # it appears below; no special regex parsing should be happening here.
    expected_error = ("Selected_features {'misspelled_feature'}"
                      " not in media_mix_model.")

    with self.assertRaisesRegex(ValueError, expected_regex=expected_error):
      plot.plot_prior_and_posterior(
          media_mix_model=mmm,
          number_of_samples_for_prior=100,
          selected_features=[
              "sigma", "intercept", "misspelled_feature"
          ])

if __name__ == "__main__":
  absltest.main()
